from sql_parser import query
from sql_parser.dml import SQLCreate
from sql_parser.ident import SQLIdentifier, SQLIdentifierPath, SQLWildcardPath
from sql_parser.query import SQLAlias, SQLNamedTable
from sql_parser.node import SQLNode, SQLNodeList
from sql_parser.query_impl import SQLField, SQLFrom, SQLJoin, SQLOrderedQuery, SQLSelect, SQLSetOp, SQLSubSelect, SQLWithSelect

from sql_parser import parse

import copy


class Refactor:

    COMMENT_COLUMN_NOT_FOUND = "[WARNING] Could not find column in knowledge: {}"
    COMMENT_TABLE_NOT_FOUND = "[WARNING] Could not find table in knowledge: {}"

    def __init__(self, knowledge:dict):
        self._knowledge = knowledge
        self.parsed = []

    def result(self):
        if self.parsed:
            sql = ""
            for parsed in self.parsed[:-1]:
                sql += '{}\n;\n\n'.format(parsed.as_sql())
            sql += self.parsed[-1].as_sql()
            return sql

    def refactor(self, sql, parse_only=False):
        self.parsed = []
        sql_commands = sql.strip('\n').split(';')
        for command in sql_commands:
            if command in ('', '\n'):
                continue
            parsed = parse(command)
            self.parsed.append(parsed)
            if not parse_only:
                self._refactor(parsed)

    def _refactor(self, parsed, tables=None):
        if isinstance(parsed, SQLWithSelect):
            self._refactor_with_select(parsed)
        elif isinstance(parsed, SQLSelect):
            self._refactor_select(parsed)
        elif isinstance(parsed, SQLFrom):
            self._refactor_from(parsed)
        elif isinstance(parsed, SQLNamedTable):
            self._refactor_named_table(parsed)
        elif isinstance(parsed, SQLJoin):
            self._refactor_join(parsed, tables)
        elif isinstance(parsed, SQLField):
            self._refactor_field(parsed, tables)
        elif isinstance(parsed, SQLIdentifierPath):
            self._refactor_identifier_path(parsed, tables)
        elif isinstance(parsed, SQLCreate):
            self._refactor_create(parsed)
        elif isinstance(parsed, SQLSetOp):
            if parsed.op in ('UNION', 'UNION ALL', 'UNION DISTINCT'):
                self._refactor_union(parsed)
            else:
                self._refactor_node(parsed, tables)
        elif isinstance(parsed, SQLNode):
            self._refactor_node(parsed, tables)

    def _refactor_with_select(self, parsed:SQLWithSelect):
        cte_tables = [table.value for table in parsed.tables]
        for i, cte in enumerate(parsed.sqls):
            self._refactor(cte)

            # Add CTE into current knowledge
            column_knowledge = {}

            select_statement = cte.select
            while True:
                if isinstance(select_statement, SQLSelect):
                    break
                if isinstance(select_statement, SQLSetOp) and select_statement.op in ('UNION', 'UNION ALL', 'UNION DISTINCT'):
                    select_statement = select_statement.left
                elif isinstance(select_statement, SQLOrderedQuery):
                    select_statement = select_statement.query
                elif isinstance(select_statement, SQLSubSelect):
                    select_statement = select_statement.query.select
                else:
                    break

            for field in select_statement.fields:
                if isinstance(field.expr, SQLWildcardPath):
                    continue
                if field.alias:
                    column_knowledge[field.alias.alias.value] = None
                else:
                    column_knowledge[field.expr.names[-1].value] = None
            self._knowledge[cte_tables[i]] = {
                'new_table' : None,
                'column_knowledge' : column_knowledge,
                'preserved' : True
            }

        self._refactor(parsed.select)

        # delete unused cte knowledge
        for cte_table in cte_tables:
            self._knowledge.pop(cte_table)


    def _refactor_select(self, parsed:SQLSelect):
        old_tables, not_found_tables = self._get_tables_and_alias(parsed.from_tables)
        
        for table in not_found_tables:
            parsed.comments.append(self.COMMENT_TABLE_NOT_FOUND.format(table))

        self._refactor(parsed.from_tables)
        
        index = 0
        while True:
            if index == len(parsed.fields):
                break
            field = parsed.fields[index]
            
            # Wildcard
            if isinstance(field.expr, SQLWildcardPath):
                if len(old_tables) == 1 or len(field.expr.names) == 0:
                    column_knowledge = self._get_column_knowledge(old_tables)
                    if len(column_knowledge) == 0:
                        index += 1
                        continue
                    # Omit columns in EXCEPT
                    except_ids = [column_id.value for column_id in field.expr.except_ids]
                    column_names = []
                    for id in column_knowledge.keys():
                        if id not in except_ids:
                            column_names.append(id)
                    
                    columns = [ 
                                SQLField(
                                    SQLIdentifierPath(
                                        SQLNodeList([SQLIdentifier(column_name)])
                                    ), None, []
                                )
                                for column_name in column_names
                            ]
                else:
                    table_alias = field.expr.names[0].value
                    table = None
                    for t, a in old_tables.items():
                        if a == table_alias:
                            table = t
                            break
                    column_knowledge = self._get_column_knowledge({table: table_alias})
                    
                    if len(column_knowledge) == 0:
                        index += 1
                        continue

                    # Omit columns in EXCEPT
                    except_ids = [column_id.value for column_id in field.expr.except_ids]
                    column_names = []
                    for id in column_knowledge.keys():
                        if id not in except_ids:
                            column_names.append(id)

                    columns = [ SQLField(
                                    SQLIdentifierPath(SQLNodeList(
                                        [SQLIdentifier(table_alias), SQLIdentifier(column_name)])
                                    ), None, []
                                )
                                for column_name in column_names
                            ]
                
                # Insert new column list from wildcard to fields
                if index == 0:
                    parsed.fields = SQLNodeList(columns) + parsed.fields[index+1:]
                else:
                    parsed.fields = parsed.fields[:index] + SQLNodeList(columns) + parsed.fields[index+1:]
                field = parsed.fields[index]

            # Identifier Path
            self._refactor(field, old_tables)
            
            index += 1

        if parsed.where_expr is not None:
            self._refactor(parsed.where_expr, old_tables)
        
    def _refactor_from(self, parsed:SQLFrom):
        if parsed is None:
            return

        tables = {}
        if isinstance(parsed.base, SQLNamedTable):
            table_id = parsed.base.table.names[0].value
            alias = None if parsed.base.alias is None else parsed.base.alias.alias.value
            tables[table_id] = alias
            self._refactor(parsed.base)
        elif isinstance(parsed.base, SQLSubSelect):
            sub_select = parsed.base
            self._refactor(sub_select)
            if sub_select.alias:
                table_alias = sub_select.alias.alias.value
                table_name = table_alias
            else:
                table_alias = '*base'
                table_name = ''
            
            # Collect column knowledge
            column_knowledge = {}

            if isinstance(sub_select.query.select, SQLSetOp) and sub_select.query.select.op in ('UNION', 'UNION ALL', 'UNION DISTINCT'):
                select_statement = sub_select.query.select.left
            elif isinstance(sub_select.query.select, SQLOrderedQuery):
                select_statement = sub_select.query.select.query
            else:
                select_statement = sub_select.query.select

            for field in select_statement.fields:
                if field.alias:
                    column_knowledge[field.alias.alias.value] = None
                else:
                    column_knowledge[field.expr.names[-1].value] = None
            additional_knowledge = {
                table_name : {
                    'new_table' : table_name,
                    'column_knowledge' : column_knowledge,
                    'preserved' : True
                }
            }
            self._knowledge = dict(self._knowledge, **additional_knowledge)

        for join_item in parsed.joins:
            if isinstance(join_item.table, SQLNamedTable):
                table_id = join_item.table.table.names[0].value
                alias = None if join_item.table.alias is None else join_item.table.alias.alias.value
                tables[table_id] = alias
            elif isinstance(join_item.table, SQLSubSelect):
                sub_select = join_item.table
                self._refactor(sub_select, tables)
                if sub_select.alias:
                    table_alias = sub_select.alias.alias.value
                    table_name = table_alias
                else:
                    table_alias = '*base'
                    table_name = ''
                
                # Collect column knowledge
                column_knowledge = {}

                if isinstance(sub_select.query.select, SQLSetOp) and sub_select.query.select.op in ('UNION', 'UNION ALL', 'UNION DISTINCT'):
                    select_statement = sub_select.query.select.left
                elif isinstance(sub_select.query.select, SQLOrderedQuery):
                    select_statement = sub_select.query.select.query
                else:
                    select_statement = sub_select.query.select
                    
                for field in select_statement.fields:
                    if isinstance(field.expr, SQLWildcardPath):
                        continue
                    if field.alias:
                        column_knowledge[field.alias.alias.value] = None
                    else:
                        column_knowledge[field.expr.names[-1].value] = None
                additional_knowledge = {
                    table_name : {
                        'new_table' : table_name,
                        'column_knowledge' : column_knowledge,
                        'preserved' : True
                    }
                }
                self._knowledge = dict(self._knowledge, **additional_knowledge)
                tables[table_name] = table_alias
            self._refactor(join_item, tables)

    def _refactor_named_table(self, parsed:SQLNamedTable):
        table_id = parsed.table.names[-1].value
        if table_id in self._knowledge.keys():
            if not self._knowledge[table_id]['preserved']:
                table_id = self._knowledge[table_id]['new_table']
        parsed.table.names[-1].value = '`' + table_id.strip('`') + '`'

    def _refactor_field(self, parsed:SQLField, tables:dict):
        if isinstance(parsed.expr, SQLIdentifierPath):
            if len(tables) == 0: # no tables found in knowledge
                return

            first_name = parsed.expr.names[0].value
            relevant_tables = tables
            
            # check whether first_name is table's alias
            for table_id, alias in tables.items():
                if alias == first_name:
                    alias = alias if alias else table_id
                    relevant_tables = {table_id : alias}
                    break
            
            column_knowledge = self._get_column_knowledge(relevant_tables)

            old_column_name = parsed.expr.names[-1].value

            if old_column_name not in column_knowledge.keys():
                parsed.comments.append(self.COMMENT_COLUMN_NOT_FOUND.format(old_column_name))
                return
                
            new_column_name_path = column_knowledge[old_column_name].split('.')
            new_column_name = new_column_name_path[-1]
            new_column_name_identifier_path = [SQLIdentifier(name) for name in new_column_name_path]
            parsed.expr.names = SQLNodeList(new_column_name_identifier_path)

            # add alias for column
            if parsed.alias is None:
                if old_column_name != new_column_name:
                    parsed.alias = SQLAlias(SQLIdentifier(old_column_name))
            elif new_column_name == parsed.alias.alias.value:
                parsed.alias = None

        else:
            self._refactor(parsed.expr, tables)

    def _refactor_join(self, parsed:SQLJoin, tables:dict):
        if isinstance(parsed.table, SQLNamedTable):
            self._refactor(parsed.table)
        if parsed.join_expr:
            self._refactor(parsed.join_expr, tables)

    def _refactor_identifier_path(self, parsed:SQLIdentifierPath, tables):
        if tables is None:
            return
        
        first_name = parsed.names[0].value
        relevant_tables = tables

        # check whether first_name is table's alias
        for table_id, alias in tables.items():
            if alias == first_name:
                alias = alias if alias else table_id
                relevant_tables = {table_id : alias}
                break
        
        column_knowledge = self._get_column_knowledge(relevant_tables)
        old_column_name = parsed.names[-1].value
        if old_column_name not in column_knowledge.keys():
            return
        new_column_name_path = column_knowledge[old_column_name].split('.')
        new_column_name_identifier_path = [SQLIdentifier(name) for name in new_column_name_path]
        parsed.names = SQLNodeList(new_column_name_identifier_path)

    def _refactor_create(self, parsed:SQLCreate):
        self._refactor(parsed.query)

        # Add CTE into current knowledge
        cte_table = parsed.table.table.names[-1].value
        column_knowledge = {}

        select_statement = parsed.query.select
        while True:
            if isinstance(select_statement, SQLSelect):
                break
            if isinstance(select_statement, SQLSetOp) and select_statement.op in ('UNION', 'UNION ALL', 'UNION DISTINCT'):
                select_statement = select_statement.left
            elif isinstance(select_statement, SQLOrderedQuery):
                select_statement = select_statement.query
            elif isinstance(select_statement, SQLSubSelect):
                select_statement = select_statement.query.select
            else:
                break

        for field in select_statement.fields:
            if field.alias:
                column_knowledge[field.alias.alias.value] = None
            else:
                column_knowledge[field.expr.names[-1].value] = None
        self._knowledge[cte_table] = {
            'new_table' : None,
            'column_knowledge' : column_knowledge,
            'preserved' : True
        }

    def _refactor_union(self, parsed:SQLSetOp):
        self._refactor(parsed.left)
        self._refactor(parsed.right)

    def _refactor_node(self, parsed:SQLNode, tables:dict=None):
        for _, val in parsed.children():
            self._refactor(val, tables)

    def _get_tables_and_alias(self, from_tables:SQLFrom) -> dict:
        tables = {}
        not_found_tables = []

        if from_tables is None:
            return tables, not_found_tables
            
        if isinstance(from_tables.base, SQLNamedTable):
            table_id = from_tables.base.table.names[-1].value
            if table_id in self._knowledge.keys():
                tables[table_id] = None if from_tables.base.alias is None \
                                    else from_tables.base.alias.alias.value
            else:
                not_found_tables.append(table_id)
        elif isinstance(from_tables.base, SQLSubSelect):
            if from_tables.base.alias:
                table_id = from_tables.base.alias.alias.value
                table_alias = table_id
            else:
                table_id = ''
                table_alias = '*base'
            tables[table_id]  = table_alias

        for join_item in from_tables.joins:
            if isinstance(join_item.table, SQLNamedTable):
                table_id = join_item.table.table.names[-1].value
                if table_id in self._knowledge.keys():
                    tables[table_id] = None if join_item.table.alias is None \
                                        else join_item.table.alias.alias.value
                else:
                    not_found_tables.append(table_id)
            elif isinstance(join_item.table, SQLSubSelect):
                if join_item.table.alias:
                    table_id = join_item.table.alias.alias.value
                    table_alias = table_id
                else:
                    table_id = ''
                    table_alias = '*base'
                tables[table_id]  = table_id

        
        not_found_tables = set(not_found_tables)
        return tables, not_found_tables

    def _get_column_knowledge(self, tables:dict):
        column_knowledge = {}
        for table, alias in tables.items():
            if table not in self._knowledge.keys():
                continue
            column_knowledge_from_table = copy.copy(self._knowledge[table]['column_knowledge'])
            if self._knowledge[table]['preserved']:
                column_knowledge_from_table = {
                                                old_column : old_column
                                                for old_column, _ in column_knowledge_from_table.items()
                                            }
            if alias and alias[0] != '*':
                    column_knowledge_from_table = {
                                                    old_column : '{}.{}'.format(alias, new_column)
                                                    for old_column, new_column in column_knowledge_from_table.items()
                                                }
            column_knowledge.update(column_knowledge_from_table)
            
        return column_knowledge



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#

import unittest
import mock

from .const import SQLConstant
from .lexer import ParsingError, SQLLexer
from .expr import SQLExpr


def mock_parse_base(lex):
    return SQLConstant(lex.consume_identifier())


@mock.patch('sql_parser.expr.SQLBaseExpr.parse',
            mock.MagicMock(side_effect=mock_parse_base))
class TestExprOp(unittest.TestCase):

    def exprt(self, v, c, r):
        # Check parsed tree against expected tree
        expr = SQLExpr.parse(SQLLexer(v))
        self.assertEqual(str(expr), c)

        # Check re-parsed generated SQL against expected tree
        v2 = expr.as_sql(True)
        expr2 = SQLExpr.parse(SQLLexer(v2))
        self.assertEqual(str(expr2), c)

        # Check non-compact re-parsed generated SQL against expected tree
        v2 = expr.as_sql(False)
        expr2 = SQLExpr.parse(SQLLexer(v2))

        # Check the representative format (get_tree)
        # TODO(scannell) bring this back!
        # self.assertEqual(parse_expression(SQLLexer(v)).get_repr(), r)

    @mock.patch('sql_parser.expr.SQLBaseExpr.parse',
                mock.MagicMock(side_effect=mock_parse_base))
    def test_bitoppexprsql(self):
        self.exprt('i>>j', 'i>>j', '>>(i,j)')
        self.exprt('i<<j', 'i<<j', '<<(i,j)')
        self.exprt('i>j', 'i>j', '>(i,j)')
        self.exprt('i<j', 'i<j', '<(i,j)')
        self.exprt('i>=j', 'i>=j', '>=(i,j)')
        self.exprt('i<=j', 'i<=j', '<=(i,j)')
        self.exprt('i!=j', 'i!=j', '!=(i,j)')
        self.exprt('i^j', 'i^j', '^(i,j)')
        self.exprt('i|j', 'i|j', '|(i,j)')
        self.exprt('i&j', 'i&j', '&(i,j)')

    def test_biopexprsql(self):
        # Test with tokenization
        self.exprt('i+j+k', 'i+j+k', '+(+(i,j),k)')
        self.exprt('i*j+k', 'i*j+k', '+(*(i,j),k)')
        self.exprt('i+j*k', 'i+j*k', '+(i,*(j,k))')

    def test_biopexpr(self):
        self.exprt('f >> g', 'f>>g', '>>(f,g)')
        self.exprt('f + g', 'f+g', '+(f,g)')
        self.exprt('f- -g', 'f- -g', '-(f,-(g))')
        self.exprt('f- ~g', 'f-~g', '-(f,~(g))')
        self.exprt('f-g', 'f-g', '-(f,g)')

    def test_logicexpr(self):
        self.exprt('A and B', 'A AND B', 'AND(A,B)')
        self.exprt('not X and Y',
                   'NOT X AND Y',
                   'AND(NOT(X),Y)')
        self.exprt('A or B and C',
                   'A OR B AND C',
                   'OR(A,AND(B,C))')
        self.exprt('A and B or C',
                   'A AND B OR C',
                   'OR(AND(A,B),C)')
    
    def test_is(self):
        self.exprt('A is null', 'A IS NULL', 'IS NULL(A)')
        self.exprt('A is not null', 'A IS NOT NULL', 'IS NOT NULL(A)')
        self.exprt('A is true', 'A IS TRUE', 'IS TRUE(A)')
        self.exprt('A is not true', 'A IS NOT TRUE', 'IS NOT TRUE(A)')
        self.exprt('A is false', 'A IS FALSE', 'IS FALSE(A)')
        self.exprt('A is not False', 'A IS NOT FALSE', 'IS NOT FALSE(A)')
        self.exprt('A is null', 'A IS NULL', 'IS NULL(A)')
        self.exprt('A is   not  /* xyz */ null',
                   'A IS NOT NULL',
                   'IS NOT NULL(A)')
        self.exprt('A like myStr', 'A LIKE myStr', 'LIKE(A,myStr)')
        self.exprt('A not like myStr',
                   'A NOT LIKE myStr',
                   'NOT LIKE(A,myStr)')
        self.exprt('a between X and Y',
                   'a BETWEEN X AND Y',
                   'BETWEEN(a,X,Y)')
        self.exprt('a NOT between X and Y',
                   'a NOT BETWEEN X AND Y',
                   'NOT BETWEEN(a,X,Y)')
        self.exprt('a IN (x,y,z)', 'a IN(x,y,z)', 'IN(a,x,y,z)')
        self.exprt('a NOT IN (x,y,z)', 'a NOT IN(x,y,z)', 'NOT IN(a,x,y,z)')

        with self.assertRaises(ParsingError):
            SQLExpr.parse(SQLLexer('a is blah'))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#

from dataclasses import dataclass
from typing import Optional

from rfmt.blocks import LineBlock as LB
from rfmt.blocks import TextBlock as TB
from rfmt.blocks import StackBlock as SB
from rfmt.blocks import ChoiceBlock as CB
from rfmt.blocks import WrapBlock as WB
from sql_parser.query_impl import SQLField

from .utils import with_commas

from .ident import SQLIdentifier, SQLIdentifierPath

from .query import SQLQuery
from .node import SQLNodeList
from .node import SQLNode
from .const import SQLConstant, SQLNumber, SQLString
from .types import SQLType
from .expr import SQLExpr


@dataclass(frozen=True)
class SQLFuncExpr(SQLExpr):
    names: SQLNodeList[SQLIdentifier]
    args: SQLNodeList[SQLExpr]

    def sqlf(self, compact):
        name = LB([x.sqlf(True) for x in self.names])
        compact_sql = LB([name, TB('(')] +
                         with_commas(True, self.args) +
                         [TB(')')])

        if compact:
            return compact_sql

        if not self.args:
            return compact_sql

        return CB([
            compact_sql,
            LB([
                LB([name, TB('(')]),
                WB(with_commas(False, self.args)),
                TB(')')
            ])
        ])


@dataclass(frozen=True)
class SQLCustomFuncs(SQLExpr):

    @staticmethod
    def consume(lex) -> 'Optional[SQLCustomFuncs]':
        # TODO(scannell) - add DATE, TIME, DATETIME, TIMESTAMP literals
        return (SQLCAST.consume(lex) or
                SQLDate.consume(lex) or
                SQLTime.consume(lex) or
                SQLCount.consume(lex) or
                SQLExists.consume(lex) or
                SQLInterval.consume(lex) or
                SQLAnalyticNavigation.consume(lex) or
                SQLExtract.consume(lex) or
                SQLCoalesce.consume(lex) or
                SQLApproxQuantiles.consume(lex) or
                SQLStructFunction.consume(lex))


@dataclass(frozen=True)
class SQLExists(SQLCustomFuncs):
    sql: SQLNode

    def sqlf(self, compact):
        return LB([
            TB('EXISTS('),
            self.sql.sqlf(True),
            TB(')'),
        ])

    @staticmethod
    def consume(lex) -> 'Optional[SQLExists]':
        if not lex.consume('EXISTS'):
            return None

        lex.expect('(')
        query = SQLQuery.parse(lex)
        lex.expect(')')

        return SQLExists(query)


@dataclass(frozen=True)
class SQLCount(SQLCustomFuncs):
    isdistinct: bool
    expr: SQLExpr

    def sqlf(self, compact):
        return LB([
            TB('COUNT(' + ('DISTINCT ' if self.isdistinct else '')),
            self.expr.sqlf(compact),
            TB(')'),
        ])

    @staticmethod
    def consume(lex) -> 'Optional[SQLCount]':
        if not lex.consume('COUNT'):
            return None
        lex.expect('(')
        isdistinct = bool(lex.consume('DISTINCT'))
        expr = SQLExpr.parse(lex)
        lex.expect(')')
        return SQLCount(isdistinct, expr)


@dataclass(frozen=True)
class SQLInterval(SQLCustomFuncs):
    sql_node: SQLNode

    def sqlf(self, compact):
        return LB([
            TB('INTERVAL('),
            self.sql_node.sqlf(compact),
            TB(')'),
        ])

    @staticmethod
    def consume(lex) -> 'Optional[SQLInterval]':
        if not lex.consume('INTERVAL'):
            return None

        if lex.consume('('):
            sql_node = (SQLString.consume(lex) or
                        lex.error('expected string'))
            lex.expect(')')
        else:
            sql_node = (SQLString.consume(lex) or
                        lex.error('expected string'))

        return SQLInterval(sql_node)


@dataclass(frozen=True)
class SQLExtract(SQLCustomFuncs):
    name: str
    part: SQLIdentifier
    expr: SQLExpr
    timezone: Optional[SQLConstant]

    def sqlf(self, compact):
        if self.timezone:
            return LB([
                TB(self.name), TB('('), self.part.sqlf(compact),
                TB(' '), TB('FROM'), TB(' '),
                self.expr.sqlf(compact),
                TB(' '), TB('AT'), TB(' '), TB('TIME'), TB(' '), TB('ZONE'), TB(' '),
                self.timezone.sqlf(compact),
                TB(')')
            ])
        else:
            return LB([
                TB(self.name), TB('('), self.part.sqlf(compact),
                TB(' '), TB('FROM'), TB(' '),
                self.expr.sqlf(compact), TB(')')
            ])

    @staticmethod
    def consume(lex) -> 'Optional[SQLExtract]':
        if not lex.consume('EXTRACT'):
            return None
        lex.expect('(')
        daypart = SQLIdentifier.parse(lex)
        lex.expect('FROM')
        date_expr = SQLExpr.parse(lex)
        timezone = None
        if lex.consume('AT'):
            lex.expect('TIME ZONE')
            timezone = SQLString.parse(lex)
        lex.expect(')')

        return SQLExtract('EXTRACT', daypart, date_expr, timezone)


@dataclass(frozen=True)
class SQLCAST(SQLCustomFuncs):
    name: str
    expr: SQLExpr
    type: SQLType

    def sqlf(self, compact):
        return LB([
            TB(self.name), TB('('), self.expr.sqlf(compact),
            TB(' '), TB('AS'), TB(' '),
            self.type.sqlf(compact), TB(')')
        ])

    @staticmethod
    def consume(lex) -> 'Optional[SQLCAST]':
        name = None
        if lex.consume('CAST'):
            name = 'CAST'
        elif lex.consume('SAFE_CAST'):
            name = 'SAFE_CAST'
        else:
            return None
        lex.expect('(')
        value_expr = SQLExpr.parse(lex)
        lex.expect('AS')
        new_type = SQLType.parse(lex)
        lex.expect(')')
        return SQLCAST(name, value_expr, new_type)


@dataclass(frozen=True)
class SQLAnalyticNavigation(SQLCustomFuncs):
    name: str
    args: SQLNodeList
    opt: str

    def sqlf(self, compact):
        if len(self.args) == 1:
            return LB([
                TB(self.name + '('),
                self.args[0].sqlf(compact),
                TB(self.opt + ')')
            ])

        return LB([
            TB(self.name + '('),
            self.args[0].sqlf(compact),
            TB(', '),
            self.args[1].sqlf(compact),
            TB(self.opt + ')')
        ])

    @staticmethod
    def consume(lex) -> 'Optional[SQLAnalyticNavigation]':
        name = (lex.consume('FIRST_VALUE') or
                lex.consume('LAST_VALUE') or
                lex.consume('NTH_VALUE') or
                lex.consume('PERCENTILE_COUNT') or
                lex.consume('PERCENTILE_DISC'))
        if not name:
            return None

        lex.expect('(')

        args = []
        while True:
            args.append(SQLExpr.parse(lex))
            if not lex.consume(','):
                break

        opt = (lex.consume('RESPECT') or
               lex.consume('IGNORE'))
        if opt:
            lex.expect('NULLS')
            opt = ' ' + opt + ' NULLS'
        elif lex.consume('NULLS'):
            opt = ' NULLS'
        else:
            opt = ''

        lex.expect(')')

        return SQLAnalyticNavigation(name, SQLNodeList(args), opt)


@dataclass(frozen=True)
class SQLDate(SQLCustomFuncs):
    name: str
    args: SQLNodeList

    def sqlf(self, compact):
        return LB([
            TB(self.name), TB('('), self.args[0].sqlf(True),
            TB(', '), TB('INTERVAL'), TB(' '),
            self.args[1].sqlf(True), TB(' '), self.args[2].sqlf(True),
            TB(')')
        ])

    @staticmethod
    def consume(lex) -> 'Optional[SQLDate]':
        name = lex.consume('DATE_ADD') or lex.consume('DATE_SUB')
        if not name:
            return None
        lex.expect('(')
        date_expr = SQLExpr.parse(lex)
        lex.expect(',')
        lex.expect('INTERVAL')
        count = SQLExpr.parse(lex)
        date_part = SQLIdentifier.parse(lex)
        lex.expect(')')
        return SQLDate(name, SQLNodeList((date_expr, count, date_part)))


@dataclass(frozen=True)
class SQLTime(SQLCustomFuncs):
    name: str
    args: SQLNodeList

    def sqlf(self, compact):
        return LB([
            TB(self.name), TB('('), self.args[0].sqlf(True),
            TB(', '), TB('INTERVAL'), TB(' '),
            self.args[1].sqlf(True), TB(' '), self.args[2].sqlf(True),
            TB(')')
        ])

    @staticmethod
    def consume(lex) -> 'Optional[SQLTime]':
        name = lex.consume('TIME_ADD') or lex.consume('TIME_SUB')  or lex.consume('TIMESTAMP_ADD')  or lex.consume('TIMESTAMP_SUB')
        if not name:
            return None
        lex.expect('(')
        date_expr = SQLExpr.parse(lex)
        lex.expect(',')
        lex.expect('INTERVAL')
        count = SQLExpr.parse(lex)
        date_part = SQLIdentifier.parse(lex)
        lex.expect(')')
        return SQLTime(name, SQLNodeList((date_expr, count, date_part)))


@dataclass(frozen=True)
class SQLCoalesce(SQLCustomFuncs):
    expr: SQLNodeList[SQLNode]

    def sqlf(self, compact):
        return LB([
            TB('COALESCE('),
            LB(with_commas(compact, self.expr)),
            TB(')'),
        ])

    @staticmethod
    def consume(lex) -> 'Optional[SQLCoalesce]':
        if not lex.consume('COALESCE'):
            return None
        lex.expect('(')

        expr = []
        while True:
            expr.append(SQLConstant.parse(lex) or SQLIdentifierPath.parse(lex))
            if not lex.consume(','):
                break
        lex.expect(')')

        expr = SQLNodeList(expr)
        return SQLCoalesce(expr)


@dataclass(frozen=True)
class SQLApproxQuantiles(SQLCustomFuncs):
    expr: SQLNode
    number: int
    offset: Optional[int]

    def sqlf(self, compact):
        if self.offset:
            return LB([
                TB('APPROX_QUANTILES('),
                self.expr.sqlf(compact), TB(','), TB(' '),
                self.number.sqlf(compact),
                TB(')'),
                TB('[OFFSET('), TB(' '),
                self.offset.sqlf(compact),
                TB(')]')
            ])
        else:
            return LB([
                TB('APPROX_QUANTILES('),
                self.expr.sqlf(compact), TB(','), TB(' '),
                self.number.sqlf(compact),
                TB(')'),
            ])

    @staticmethod
    def consume(lex) -> 'Optional[SQLApproxQuantiles]':
        if not lex.consume('APPROX_QUANTILES'):
            return None
        lex.expect('(')

        expr = SQLIdentifierPath.parse(lex)

        lex.expect(',')
        number = SQLNumber.parse(lex)

        lex.expect(')')

        offset = None
        if lex.consume('['):
            lex.expect('OFFSET')
            lex.consume('(')
            offset = SQLNumber.parse(lex)
            lex.consume(')')
            lex.expect(']')

        return SQLApproxQuantiles(expr, number, offset)

@dataclass(frozen=True)
class SQLStructFunction(SQLCustomFuncs):
    expr: SQLNodeList[SQLNode]

    def sqlf(self, compact):
        return LB([
            TB('STRUCT('),
            SB(with_commas(compact, self.expr)),
            TB(')'),
        ])

    @staticmethod
    def consume(lex) -> 'Optional[SQLStructFunction]':
        if not lex.consume('STRUCT'):
            return None
            
        lex.expect('(')
        expr = []
        while True:
            expr.append(SQLField.parse(lex))
            if not lex.consume(','):
                break
        lex.expect(')')

        expr = SQLNodeList(expr)
        return SQLStructFunction(expr)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

"""SQL Lexer - parse basic constructs from a StringReader.

"""

from dataclasses import dataclass
from typing import Any
from typing import Optional
from typing import Union

from rfmt.blocks import TextBlock as TB

from .expr import SQLExpr


@dataclass(frozen=True)
class SQLConstant(SQLExpr):
    value: Any = None
    """SQLConstant - basic types.

    Attributes:
      val: Python value type
    """

    def sqlf(self, compact):
        del compact  # Unused
        return TB(str(self.value))

    @staticmethod
    def consume(lex) -> 'Optional[SQLConstant]':
        return (SQLNull.consume(lex) or
                SQLBool.consume(lex) or
                SQLNumber.consume(lex) or
                SQLString.consume(lex))


@dataclass(frozen=True)
class SQLNull(SQLConstant):
    """SQLNull - NULL constant.
    """
    def sqlf(self, compact):
        del compact  # Unused
        return TB('NULL')

    @staticmethod
    def consume(lex) -> 'Optional[SQLNull]':
        if lex.consume('NULL'):
            return SQLNull()
        return None


@dataclass(frozen=True)
class SQLNumber(SQLConstant):
    value: Union[float, int]

    @staticmethod
    def consume(lex) -> 'Optional[SQLNumber]':
        num = lex.consume_number()
        if num is None:
            return None

        return SQLNumber(num)


@dataclass(frozen=True)
class SQLBool(SQLConstant):
    value: bool
    """SQLBool - basic types.

    Attributes:
      val: Python true/false
    """
    def sqlf(self, compact):
        del compact  # Unused
        return TB(str(self.value).upper())

    @staticmethod
    def consume(lex) -> 'Optional[SQLBool]':
        if lex.consume('TRUE'):
            return SQLBool(True)
        if lex.consume('FALSE'):
            return SQLBool(False)
        return None


@dataclass(frozen=True)
class SQLString(SQLConstant):
    value: str
    flag: Optional[str] = None
    """SQLString - basic string.

    Attributes:
      val: Python string
    """

    ESCAPE_CHARS = ["\\", "'"]

    def sqlf(self, compact):
        del compact  # Unused

        # Escape characters
        out_string = self.value
        for escape_char in SQLString.ESCAPE_CHARS:
            out_string = out_string.replace(escape_char, "\\" + escape_char)

        return TB('{}\'{}\''.format(
            self.flag or '',
            out_string,
        ))

    @staticmethod
    def consume(lex) -> 'Optional[SQLString]':
        sql_str = lex.consume_string()
        if not sql_str:
            return None

        return SQLString(sql_str[0], sql_str[1])



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#

import unittest
import mock

from .const import SQLConstant
from .lexer import ParsingError, SQLLexer
from .expr import SQLExpr


def mock_type_parser(lex):
    lex.expect('TYPE')
    return SQLConstant('TYPE')


@mock.patch('sql_parser.expr_funcs.SQLType.parse',
            mock.MagicMock(side_effect=mock_type_parser))
@mock.patch('sql_parser.expr_op.SQLType.parse',
            mock.MagicMock(side_effect=mock_type_parser))
class TestExpr(unittest.TestCase):

    def exprt(self, val, check):
        # Check parsed tree against expected tree
        expr = SQLExpr.parse(SQLLexer(val))
        self.assertEqual(str(expr), check)

        # Check re-parsed generated SQL against expected tree
        val2 = expr.as_sql(True)
        expr2 = SQLExpr.parse(SQLLexer(val2))
        self.assertEqual(str(expr2), check)

        # Check non-compact re-parsed generated SQL against expected tree
        val3 = expr.as_sql(False)
        expr3 = SQLExpr.parse(SQLLexer(val3))
        self.assertEqual(str(expr3), check)

    def test_bitoppexprsql(self):
        self.exprt('i>>j', 'i>>j')
        self.exprt('i<<j', 'i<<j')
        self.exprt('i>j', 'i>j')
        self.exprt('i<j', 'i<j')
        self.exprt('i>=j', 'i>=j')
        self.exprt('i<=j', 'i<=j')
        self.exprt('i!=j', 'i!=j')
        self.exprt('i^j', 'i^j')
        self.exprt('i|j', 'i|j')
        self.exprt('i&j', 'i&j')

    def test_biopexprsql(self):
        # Test with tokenization
        self.exprt('i+j+k', 'i+j+k')
        self.exprt('i*j+k', 'i*j+k')
        self.exprt('i+j*k', 'i+j*k')
        self.exprt('i+(j/k)', 'i+(j/k)')
        self.exprt('(i+(j*k))', '(i+(j*k))')
        self.exprt('i*(j+k)', 'i*(j+k)')
        self.exprt('i+(j*l+k)', 'i+(j*l+k)')

    def test_biopexpr(self):
        self.exprt('f >> g', 'f>>g')
        self.exprt('f + g', 'f+g')
        self.exprt('f- -g', 'f- -g')
        self.exprt('f- ~g', 'f-~g')
        self.exprt('(a+b)*c', '(a+b)*c')
        self.exprt('f-g', 'f-g')
        self.exprt('F(a)-b', 'F(a)-b')
        self.exprt('F(a)- -b', 'F(a)- -b')

    def test_logicexpr(self):
        self.exprt('A and B', 'A AND B')
        self.exprt('not X and Y',
                   'NOT X AND Y')
        self.exprt('A or B and C',
                   'A OR B AND C')
        self.exprt('A and B or C',
                   'A AND B OR C')

    def test_is(self):
        self.exprt('A is null', 'A IS NULL')
        self.exprt('A is not null', 'A IS NOT NULL')
        self.exprt('A is true', 'A IS TRUE')
        self.exprt('A is not true', 'A IS NOT TRUE')
        self.exprt('A is false', 'A IS FALSE')
        self.exprt('A is not False', 'A IS NOT FALSE')
        self.exprt('A is null', 'A IS NULL')
        self.exprt('A is   not  /* xyz */ null',
                   'A IS NOT NULL')
        self.exprt('A like myStr', 'A LIKE myStr')
        self.exprt('A not like myStr',
                   'A NOT LIKE myStr')
        self.exprt('a between X and Y',
                   'a BETWEEN X AND Y')
        self.exprt('a NOT between X and Y',
                   'a NOT BETWEEN X AND Y')
        self.exprt('a IN (x,y,z)', 'a IN(x,y,z)')
        self.exprt('a NOT IN (x,y,z)', 'a NOT IN(x,y,z)')

        with self.assertRaises(ParsingError):
            SQLExpr.parse(SQLLexer('a is blah'))

    def test_func(self):
        self.exprt('funcA(a,b)', 'funcA(a,b)')
        self.exprt('funcA(x - i) ', 'funcA(x-i)')

    def test_special(self):
        # Date
        self.exprt('date_Add(myDateVar, INTERVAL a DAY)',
                   'DATE_ADD(myDateVar,INTERVAL a DAY)')
        self.exprt('date_sub(myDateVar, INTERVAL b QUARTER)',
                   'DATE_SUB(myDateVar,INTERVAL b QUARTER)')

        # Time
        self.exprt('TIME_ADD(a, INTERVAL 5 minute)',
                    'TIME_ADD(a,INTERVAL 5 minute)')
        self.exprt('TIME_SUB(a, INTERVAL b second)',
                    'TIME_SUB(a,INTERVAL b second)')

        # Cast
        self.exprt('cast(myVar AS TYPE)',
                   'CAST(myVar AS TYPE)')
        self.exprt('safe_cast(yourVar AS TYPE)',
                   'SAFE_CAST(yourVar AS TYPE)')

        # Colon Cast
        self.exprt('myVar::type', 'myVar::TYPE')

        # CASE
        self.exprt('case when a then b end',
                   'CASE WHEN a THEN b END')
        self.exprt('case when x then y end',
                   'CASE WHEN x THEN y END')
        self.exprt('case when a.b=f then p end',
                   'CASE WHEN a.b=f THEN p END')
        self.exprt('case when i then j else k end',
                   'CASE WHEN i THEN j ELSE k END')
        self.exprt('case F WHEN G THEN H end',
                   'CASE F WHEN G THEN H END')

        # Extract
        self.exprt('extract(minute from MyDate)',
                   'EXTRACT(minute FROM MyDate)')
        self.exprt('extract(minute from MyDate AT TIME ZONE "America/Los_Angeles")',
                   'EXTRACT(minute FROM MyDate AT TIME ZONE \'America/Los_Angeles\')')
        # Count
        self.exprt('count(a)', 'COUNT(a)')
        self.exprt('count(distinct a)', 'COUNT(DISTINCT a)')

        # Interval
        self.exprt('INTERVAL \'1 second\'',
                   'INTERVAL(\'1 second\')')
        self.exprt('INTERVAL(\'1 second\')',
                   'INTERVAL(\'1 second\')')

        # Approx Quantiles
        self.exprt('APPROX_QUANTILES(x, 100)',
                   'APPROX_QUANTILES(x,100)')
        # Array AGG
        # TODO(scannell)

    def test_in(self):
        self.exprt('a IN (SELECT 1)', 'a IN(SELECT 1)')
        self.exprt('(a,b) IN (SELECT 1)', '(a,b)IN(SELECT 1)')

    def test_analytic_navigation(self):
        self.exprt('last_value(a ignore nulls)',
                   'LAST_VALUE(a IGNORE NULLS)')

    def test_partition_over(self):
        self.exprt('last_value(x ignore nulls) over ()',
                   'LAST_VALUE(x IGNORE NULLS)OVER()')

    def test_exists(self):
        self.exprt('exists(select 1)',
                   'EXISTS(SELECT 1)')

    def test_array_agg(self):
        self.exprt('ARRAY_AGG(DISTINCT   x  LIMIT  1)',
                   'ARRAY_AGG(DISTINCT x LIMIT 1)')

    def test_string_agg(self):
        self.exprt('STRING_AGG(DISTINCT   x )',
                   'STRING_AGG(DISTINCT x)')

    def test_array_lit(self):
        self.exprt('[1,2,3,4]', '[1,2,3,4]')

    def test_struct(self):
        self.exprt('("a", 1)', '(\'a\',1)')

    def test_array_select(self):
        self.exprt('ARRAY(SELECT x FROM abc)',
                   'ARRAY(SELECT x FROM abc)')

    def test_coalesce(self):
        self.exprt('COALESCE(a, b, c, 0)',
                    'COALESCE(a,b,c,0)')



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#

"""SQL Types.

Parsing and AST Nodes
"""

from dataclasses import dataclass
from typing import Optional
from typing import List

from rfmt.blocks import LineBlock as LB
from rfmt.blocks import TextBlock as TB
from rfmt.blocks import WrapBlock as WB

from .node import SQLNode
from .node import SQLNodeList

from .const import SQLNumber
from .ident import SQLIdentifier


@dataclass(frozen=True)
class SQLType(SQLNode):

    @staticmethod
    def parse(lex) -> 'SQLType':
        return (SQLType.consume(lex) or
                lex.error('expected type'))

    @staticmethod
    def consume(lex) -> 'Optional[SQLType]':
        return (SQLConcreteType.consume(lex) or
                SQLVarchar.consume(lex) or
                SQLArray.consume(lex) or
                SQLStruct.consume(lex))


@dataclass(frozen=True)
class SQLConcreteType(SQLType):
    """Concrete SQLConcreteType.
    """
    value: str

    def sqlf(self, compact):
        del compact  # Unused
        return TB(self.value)

    @staticmethod
    def consume(lex) -> 'Optional[SQLConcreteType]':
        typ = lex.consume_any(['INT64', 'FLOAT64', 'BOOL', 'BYTES', 'DATE',
                               'DATETIME', 'TIME', 'TIMESTAMP', 'STRING',
                               'BIGINT', 'SMALLINT', 'INT4', 'ABSTIME',
                               'INTEGER', 'BOOLEAN', 'NUMERIC'])
        if not typ:
            return None

        return SQLConcreteType(typ)


@dataclass(frozen=True)
class SQLNamedType(SQLType):
    """Named SQL Type.
    """
    tname: SQLIdentifier
    type: Optional[SQLType]

    def sqlf(self, compact):
        return LB([
            self.tname.sqlf(compact),
            TB(' '),
            self.type.sqlf(compact)
        ])


@dataclass(frozen=True)
class SQLArray(SQLType):
    """SQL Array.
    """
    atype: SQLType

    def sqlf(self, compact):
        return LB([
            TB('ARRAY<'),
            self.atype.sqlf(compact),
            TB('>')
        ])

    @staticmethod
    def consume(lex) -> 'Optional[SQLArray]':
        if not lex.consume('ARRAY'):
            return None

        lex.expect('<')
        ntype = SQLType.parse(lex)
        lex.expect('>')

        return SQLArray(ntype)


@dataclass(frozen=True)
class SQLVarchar(SQLType):
    size: SQLNode

    def sqlf(self, compact):
        return LB([
            TB('VARCHAR('),
            self.size.sqlf(True),
            TB(')')
        ])

    @staticmethod
    def consume(lex) -> 'Optional[SQLVarchar]':
        if not lex.consume('VARCHAR'):
            return None

        lex.expect('(')
        size = (SQLNumber.consume(lex) or
                lex.error('expected number'))
        lex.expect(')')

        return SQLVarchar(size)


@dataclass(frozen=True)
class SQLStruct(SQLType):
    """SQL Structure.
    """
    vals: SQLNodeList[SQLType]

    def sqlf(self, compact):
        sep = ','
        if compact:
            sep += ' '

        fields = []
        for field in self.vals[0:-1]:
            fields.append(LB([
                field.sqlf(True), TB(sep)
            ]))
        fields.append(self.vals[-1].sqlf(True))

        if compact:
            return LB(
                [TB('STRUCT<')] +
                fields +
                [TB('>')]
            )
        return LB([
            TB('STRUCT<'),
            WB(fields),
            TB('>')
        ])

    @staticmethod
    def consume(lex) -> 'Optional[SQLStruct]':
        if not lex.consume('STRUCT'):
            return None

        lex.expect('<')
        types: List[SQLType] = []
        while True:
            ptype = SQLType.consume(lex)
            if ptype is None:
                name = SQLIdentifier.parse(lex)
                types.append(SQLNamedType(name, SQLType.parse(lex)))
            else:
                types.append(ptype)
            if not lex.consume(','):
                break
        lex.expect('>')
        return SQLStruct(SQLNodeList(types))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


from dataclasses import dataclass
from typing import Optional

from rfmt.blocks import LineBlock as LB
from rfmt.blocks import TextBlock as TB

from .node import SQLNode, SQLNodeList

from .ident import SQLIdentifierPath
from .ident import SQLIdentifier


@dataclass(frozen=True)
class SQLAlias(SQLNode):
    alias: SQLIdentifier

    def sqlf(self, compact):
        return LB([TB('AS '), self.alias.sqlf(True)])

    @staticmethod
    def consume(lex) -> 'Optional[SQLAlias]':
        if lex.consume('AS'):
            return SQLAlias(SQLIdentifier.parse(lex))
        alias = SQLIdentifier.consume(lex)
        if not alias:
            return None
        return SQLAlias(alias)


@dataclass(frozen=False)
class SQLTableSource(SQLNode):

    @staticmethod
    def parse(lex) -> 'SQLTableSource':
        from .query_impl import SQLSubSelect
        from .expr_funcs import SQLFuncExpr

        if lex.consume('UNNEST'):
            lex.expect('(')
            name = SQLIdentifier('UNNEST')
            arg = SQLFuncExpr.parse(lex) or SQLIdentifierPath.parse(lex) 
            lex.expect(')')
            alias = SQLAlias.consume(lex)
            expr = SQLFuncExpr(SQLNodeList([name]), SQLNodeList([arg]))
            return SQLFunctionTable(expr, alias)
        
        if lex.consume('('):
            table = SQLQuery.parse(lex)
            lex.expect(')')
            alias = SQLAlias.consume(lex)
            return SQLSubSelect(table, alias)

        # Named Table (read-context)
        return SQLNamedTable.parse(lex, False)


@dataclass(frozen=False)
class SQLNamedTable(SQLTableSource):
    table: SQLIdentifierPath
    alias: Optional[SQLAlias] = None
    is_write: bool = False

    def sqlf(self, compact):
        expr = self.table.sqlf(compact)
        if self.alias is not None:
            expr = LB(
                [expr, TB(' '), self.alias.sqlf(True)])
        return expr

    @staticmethod
    def parse(lex, is_write=False) -> 'SQLNamedTable':
        table_name = SQLIdentifierPath.parse(lex)
        alias = SQLAlias.consume(lex)
        return SQLNamedTable(table_name, alias, is_write)

    @staticmethod
    def parse_no_alias(lex, is_write=False) -> 'SQLNamedTable':
        table_name = SQLIdentifierPath.parse(lex)
        return SQLNamedTable(table_name, None, is_write)


@dataclass(frozen=True)
class SQLQuery(SQLTableSource):

    @staticmethod
    def parse(lex) -> 'SQLQuery':
        from .query_impl import SQLWithSelect
        return SQLWithSelect.parse(lex)

    @staticmethod
    def consume(lex) -> 'Optional[SQLQuery]':
        from .query_impl import SQLWithSelect
        return SQLWithSelect.consume(lex)

@dataclass(frozen=True)
class SQLFunctionTable(SQLTableSource):
    from .expr import SQLExpr

    expr: SQLExpr
    alias: Optional[SQLAlias] = None
    is_write: bool = False

    def sqlf(self, compact):
        expr = self.expr.sqlf(compact)
        if self.alias is not None:
            expr = LB(
                [expr, TB(' '), self.alias.sqlf(True)])
        return expr


# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#

"""SQL Lexer - parse basic constructs from a StringReader.

"""

from dataclasses import dataclass
from typing import List
from typing import Optional

from rfmt.blocks import TextBlock as TB
from rfmt.blocks import WrapBlock as WB
from rfmt.blocks import LineBlock as LB

from .node import SQLNodeList
from .node import SQLNode
from .expr import SQLExpr


@dataclass(frozen=False)
class SQLIdentifier(SQLNode):
    value: str

    def sqlf(self, compact):
        del compact
        return TB(self.value)

    @staticmethod
    def consume(lex) -> 'Optional[SQLIdentifier]':
        word = lex.consume_identifier()
        if not word:
            return None
        return SQLIdentifier(word)

    @staticmethod
    def parse(lex) -> 'SQLIdentifier':
        return (SQLIdentifier.consume(lex) or
                lex.error('expected identifier'))


@dataclass(frozen=False)
class SQLIdentifierPath(SQLExpr):
    names: SQLNodeList[SQLIdentifier]

    def sqlf(self, compact):
        del compact  # Unused
        words = []
        for i in range(len(self.names)):
            words.append(self.names[i].sqlf(True))
            if i < (len(self.names)-1):
                words.append(TB('.'))
        return LB(words)

    @staticmethod
    def parse(lex) -> 'SQLIdentifierPath':
        names: List[SQLIdentifier] = []

        while True:
            if lex.consume('*'):
                ids: List[SQLIdentifier] = []
                if lex.consume('EXCEPT'):
                    lex.expect('(')
                    while True:
                        ids.append(SQLIdentifier.parse(lex))
                        if not lex.consume(','):
                            break
                    lex.expect(')')
                return SQLWildcardPath(SQLNodeList(names), SQLNodeList(ids))

            names.append(SQLIdentifier.parse(lex))

            if not lex.consume('.'):
                break

        return SQLIdentifierPath(SQLNodeList(names))


@dataclass(frozen=False)
class SQLWildcardPath(SQLIdentifierPath):
    names: SQLNodeList[SQLIdentifier]
    except_ids: SQLNodeList[SQLIdentifier]

    def sqlf(self, compact):
        words = []
        if self.names:
            words.append(TB('.'.join([name.value for name in self.names]) + '.*'))
        else:
            words.append(TB('*'))
        if self.except_ids:
            words.append(TB(' '))
            words.append(TB('EXCEPT('))
            except_list = [TB(x.value + ',') for x in self.except_ids[:-1]]
            except_list.append(TB(self.except_ids[-1].value))
            words.append(WB(except_list))
            words.append(TB(')'))
        return LB(words)



import setuptools

with open("sql_refactor/README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="sql-refactor",
    version="0.0.6",
    author="Muhamad Tohir",
    author_email="maztohir@gmail.com",
    description="Transform column and table name to desired one",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/maztohir/sample-sql-translator",
    project_urls={
        "Bug Tracker": "https://github.com/maztohir/sample-sql-translator/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=['sql_refactor', 'sql_parser', 'rfmt'],
    python_requires=">=3.7.9",
)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


from .parser import SQLScript
from .lexer import SQLLexer


def parse(sql):
    return SQLScript.parse(SQLLexer(sql))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#

import unittest

from .const import SQLConstant
from .lexer import SQLLexer


class TestLexer(unittest.TestCase):

    def check_const(self, ival, sql):
        lex = SQLLexer(ival)
        val = SQLConstant.consume(lex)
        self.assertEqual(val.as_sql(True), sql)
        self.assertEqual(val.as_sql(False), sql)
        self.assertEqual(str(val), sql)
        self.assertEqual(SQLConstant.consume(lex), None)

    def test_float(self):
        self.check_const('1234.42', '1234.42')

    def test_integer(self):
        self.check_const('1234', '1234')

    def test_quote(self):
        self.check_const('\'asdf\'', '\'asdf\'')
        self.check_const('"asdf"', '\'asdf\'')
        self.check_const('"""asdf"""', '\'asdf\'')

    def test_flag(self):
        self.check_const('b\'asdf\'', 'b\'asdf\'')

    def test_constant_type(self):
        self.check_const('true', 'TRUE')
        self.check_const('false', 'FALSE')
        self.check_const('null', 'NULL')



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Parser generator for SQL Commands independent data operations.

  Usage:
    cmd = SQLCommand.consume(lex)

"""

from dataclasses import dataclass
from typing import Optional

from rfmt.blocks import LineBlock as LB
from rfmt.blocks import TextBlock as TB

from .query import SQLNamedTable
from .node import SQLNode


@dataclass(frozen=True)
class SQLCommand(SQLNode):

    @staticmethod
    def consume(lex) -> 'Optional[SQLCommand]':
        return (SQLTruncate.consume(lex) or
                SQLGenerateStatistics.consume(lex) or
                SQLGroom.consume(lex) or
                SQLTrans.consume(lex))


@dataclass(frozen=True)
class SQLTruncate(SQLCommand):
    table: SQLNamedTable

    def sqlf(self, compact):
        return LB([
            TB('TRUNCATE TABLE '),
            self.table.sqlf(compact=True),
            TB(';'),
        ])

    @staticmethod
    def consume(lex) -> 'Optional[SQLTruncate]':
        if not lex.consume('TRUNCATE'):
            return None

        lex.consume('TABLE')

        table_name = SQLNamedTable.parse(lex, is_write=True)

        return SQLTruncate(table_name)


@dataclass(frozen=True)
class SQLGenerateStatistics(SQLCommand):
    table: SQLNamedTable

    def sqlf(self, compact):
        return LB([
            TB('GENERATE STATISTICS ON '),
            self.table.sqlf(compact=True),
            TB(';'),
        ])

    @staticmethod
    def consume(lex) -> 'Optional[SQLGenerateStatistics]':
        if not lex.consume(['GENERATE', 'STATISTICS', 'ON']):
            return None

        lex.consume('TABLE')
        table = SQLNamedTable.parse(lex, is_write=True)

        return SQLGenerateStatistics(table)


@dataclass(frozen=True)
class SQLGroom(SQLCommand):
    table: SQLNamedTable

    def sqlf(self, compact):
        return LB([
            TB('GROOM TABLE '),
            self.table.sqlf(compact=True),
            TB(' RECORDS ALL;'),
        ])

    @staticmethod
    def consume(lex) -> 'Optional[SQLGroom]':
        if not lex.consume(['GROOM', 'TABLE']):
            return None

        table = SQLNamedTable.parse(lex, is_write=True)

        lex.expect('RECORDS')
        lex.expect('ALL')

        return SQLGroom(table)


@dataclass(frozen=True)
class SQLTrans(SQLCommand):
    trans: str

    def sqlf(self, compact):
        return TB(self.trans + ';')

    @staticmethod
    def consume(lex) -> 'Optional[SQLTrans]':
        if lex.consume('BEGIN'):
            return SQLTrans('BEGIN')
        if lex.consume('COMMIT'):
            return SQLTrans('COMMIT')
        if lex.consume('ABORT'):
            return SQLTrans('ABORT')
        return None
